/*************************** Little Othello *********************************
 * Projet Othello DEUG ][ Technologie Industrielle                          *
 *   UHA/FST - Mulhouse - France                                            *
 *   by Andre Rassat & Matthieu Klein 1996 (v1.1Us Dos version)               *
 *   e-mail: matthieu@hrnet.fr or 3614 code 'rtel1' BAL DCC                 *
 * The TRIPLEX font has been converted to an .OBJ file and integrated into  *
 * graphic.lib. The same goes for the graphic pilot EGAVGA.OBJ. Then please *
 * refer to UTIL.DOC, included with Turbo C++ 1.01 in order to do it before *
 * compiling (or else you won't be able to compile...)                      *
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <mem.h>
#include <graphics.h>
#include <alloc.h>  // used for save_load(int)
#include <dir.h>    // used for save_load(int)
#include <string.h> // used for save_load(int)
#include <dos.h>  // in order to use int86 (mouse etc...)

/********************************* macros ***********************************/

#define TRUE 1
#define FALSE 0

#define Max_note 20000

// compute level to depth
#define Novice(j1,j2) (((64-j1-j2)<=9)?64-j1-j2:3)
#define Meduim(j1,j2) (((64-j1-j2)<=13)?64-j1-j2:4)
#define Expert(j1,j2) (((64-j1-j2)<=15)?64-j1-j2:5)
#define User(player)  (status.joueur[player].user)

// move structure decoding
#define Mask_x(X) ((X & 224) >> 5) + 1    // Extract a move pos. X
#define Mask_y(Y) ((Y & 28) >> 2) + 1     // Extract a move pos. Y
#define Mask_c(C) (C & 3)                 // Extract a move color

#define Reverse(J) (-J+3)                 // BLACK <-> WHITE  ( 1 <-> 2 )

#define Waitforretrace()   while(!(inport(0x3DA) & 8))
/* Pretty good macro. Basically,
   it uses the Input Status #1 register on the VGA card, and tests
   BIT 3 (that is the & 8) for the vertical retrace mode */

#define Erase_KeybBuff while (Regs.h.ah = 0x01, int86(0x16, &Regs, &Regs), \
                              Regs.h.ah = 0x00, !(Regs.x.flags & 64))      \
                              int86(0x16, &Regs, &Regs);
// I can't find equivalent C function, so I had to make it ASM.
// fflush(stdin); is not for keyboard buffer...

// #define INTR 0x1C

#define COUL_TEXT    BLACK       // 0
#define COUL_RECT    LIGHTBLUE   // 9
#define LINE1        DARKGRAY    // 8
#define LINE2        CYAN        // 3

/***************************** mouse functions ******************************/

void mouse_win (int, int, int, int);
void mouse_curs (int);
int mouse_bout (void);
void mouse_pos (int*, int*);
void mouse_set (int, int);
void gest_sour (int*, int*);

/****************************** board functions  ****************************/

void presentation (void);
void init_game();
void fleches(int, int);
void couleur_courante(int);
void pixel2case (int, int, int*, int*);
void dessine_matrice_damier();
void apply_pion (int, int, int, int);
void textxy (int, int, char*, int);
void bouton (int, int, int, char*);
void refresh_buttons(void);
void save_load(int);
void message();
// void interrupt (*oldhandler)(void);
// void interrupt display_node(void);

/***************************** engine functions *****************************/

typedef unsigned char octet;

void pass();
void backward_forward();
void apply_move(octet new_board[10][10], octet* move, int mode);
int coup_possibles(octet board[10][10] ,octet possibles[33], octet couleur);
int evaluation_damier(octet le_damier[10][10]);
int alpha_beta(octet board[10][10], octet depth, octet joueur, octet* move, int* score, int note, octet who);
int minimax(octet board[10][10], octet depth, octet joueur, octet* move, int* score);


/*************************** variables globales *****************************/

octet damier[10][10]; // damier[0][0] -> damier[9][9] = 100 octets

octet historique[63]; // Backward/Forward moves
// historique[62] is used for swap (2*2) pieces at each NEW GAME

// *historique EQU historique[0] contains the number of moves in the
// table. historique[1] is a pseudo pointer which points in the table
// the move which has been played or removed, ok ?

unsigned long count = 0; // Alpha-Beta nodes counter

signed int perimetre[16] = {0,-1,1,-1,1,0,1,1,0,1,-1,1,-1,0,-1,-1};

//      this table gives the relative x,y coordinates of the immediate
//      perimeter of a board's square. ex: perimetre[0] and perimetre[1] give
//      (x,y)=(0,-1), which, when added to a board's square, give the
//      immediate above square coordinates

struct { int current,             // value: 0 (NEWGAME), 1 (BLACK), 2 (WHITE)
             fin,                 // End game because won. TRUE or FALSE
             pause,               // used for 'PAUSE' mode,
             hint_level,          // depth of hint moves
             algo;                // FALSE = Alpha-beta; TRUE = Min-Max
         struct zz { int humain,  // TRUE = human, FALSE = Pc
                         niveau,  // if humain = FALSE, then niveau = Depth
                         user;    // Pc depth control (1 to 60)
                   } joueur[2];   // used with status.current
       } status;

union REGS Regs;  // registers i/o access (interrupt)

// flag_info used for numerical informations' management (mob. + nmbr)
octet flag_info, flag_dialogue;

/****************************************************************************
 *                               ... MAIN ...                               *
 ****************************************************************************/

void main()

{ octet coup, possibles[33] = { 0 };

  int x = 0, y,
      note,
      i, j, j1, j2, k; // rem: black = j1 / white = j2

  //  char *j1c, *j2c;
  char j1c[20], j2c[20];

  presentation();  // intro page + call init_game();

  do { gest_sour( &x, &y);

       if (flag_info) // display the numerical info Pieces/Mobility
         { Waitforretrace();
           setfillstyle(SOLID_FILL,BLACK);
           bar(130,350,275,397);
           // j1c=(char*)malloc(3);
           // j2c=(char*)malloc(3);
           itoa(coup_possibles(damier,possibles,1),j1c,10);
           itoa(coup_possibles(damier,possibles,2),j2c,10);
           setcolor(LIGHTBLUE);
           outtextxy(168,375,j1c); // player 1's mobility
           outtextxy(229,375,j2c); // player 2's mobility
           j1 = j2 = 0;
           for (j=1; j<=8; j++)
             for (i=1; i<=8; i++)
               if (damier[i][j]) if (damier[i][j]==1)
                 j1++;
               else j2++;
           itoa(j1,j1c,10); itoa(j2,j2c,10);
           outtextxy(168,360,j1c);    // player 1 (BLACK) 's nbr of pieces
           outtextxy(229,360,j2c);    // player 2 (WHITE) 's nbr of pieces
           // free(j1c); free(j2c);

           //// If humain, erase Depth + Board eval. + Tree nodes
           for (note=0; note<=1; note++)
             { if (status.joueur[note].humain)
                 { bar(note?197:130,397,note?275:196,428);
                   for (j=390; j<=420; j+=15)
                     outtextxy(note?229:168,j,"---");
                 }
               else // display depth
                 { i = status.joueur[note].niveau;
                   itoa((i==1)?Novice(j1,j2):(i==2)?Meduim(j1,j2):
                        (i==3)?Expert(j1,j2):User(note),j1c,10);
                   outtextxy(note?229:168,390,j1c);
                 }
             }
           flag_info = FALSE;
         }

       if (status.current && !status.pause)
         if (coup_possibles(damier, possibles, status.current))
           if (status.joueur[status.current-1].humain)
             { if (x) { coup = ((x-1)<<5)|((y-1)<<2)|status.current;
                        // x = 0: click between 2 cases (cf 'pixel2case')
                        // or reinit at the end of the main()'s while
                        for (i=1; i<=*possibles; i++)
                          if (possibles[i]==coup)
                            { apply_move(damier, &coup, TRUE);
                              // TRUE to record in the historical record
                              couleur_courante(Reverse(status.current));
                            }
                      }
             }
           else   // L-Othello plays  !!
	     { j = status.joueur[status.current-1].niveau;
	       j1 = (j==1)?Novice(j1,j2):(j==2)?Meduim(j1,j2):
		    (j==3)?Expert(j1,j2):User(status.current-1);
               count = 0;
               // oldhandler = getvect(INTR);
               // setvect(INTR, display_node);
               while(i=(status.algo?minimax(damier,
                                            j1,
                                            status.current,
                                            &coup,
                                            &note)
                                    :alpha_beta(damier,
                                               j1,
                                               status.current,
                                               &coup,
                                               &note,
                                               0, 0)),
		    (j!=4)?(j1>6)?j1=6:j1--:j1--,
                    (((note==Max_note && status.current==2)||
                       (note==-Max_note &&  status.current==1)
                      ) && j1 && i) );
                                          // if TRUE, the other player
                                          // can WIN !! Lothello tries to
                                          // maximize the score, therefore
                                          // the other player (if human!!)
                                          // may lose.
                                          // Here the while function is used
                                          // only for one turn.
                 if (i)
                 { itoa(note,j1c,10); // (j1c for WHITE/BLACK)
                   ltoa(count,j2c,10);
                   setfillstyle(SOLID_FILL,BLACK);
                   setcolor(LIGHTBLUE);
                   if (status.current==1)    // BLACK player zone
                     // displays Board Evaluation + counter
                     { bar(130,397,196,428);
                       outtextxy(168,405,j1c);
                       outtextxy(168,420,j2c);
                     }
                   else                      // WHITE player zone
                     { bar(197,397,275,428);
                       outtextxy(229,405,j1c);
                       outtextxy(229,420,j2c);
                     }
                   apply_move(damier, &coup, TRUE);
                   couleur_courante(Reverse(status.current));
                 } // End if alpha-beta
               // setvect(INTR, oldhandler);
             } // End 'L-Othello play'
         else if (coup_possibles(damier, possibles, Reverse(status.current)))
                if (status.joueur[status.current-1].humain)
                  { if (flag_dialogue) // case human has to pass his turn
                      { setfillstyle(SOLID_FILL,COUL_RECT);
                        bar(282,32,598,58); // erase the dialog box's txt
                        settextstyle(DEFAULT_FONT,HORIZ_DIR,1);
                        textxy(440,39,(status.current-1)?"NOTE: WHITE PLAYER HAS TO PASS HIS TURN":
                        "NOTE: BLACK PLAYER HAS TO PASS HIS TURN",COUL_TEXT);
                        textxy(440,51,evaluation_damier(damier)>0?(status.current-1)?
                        "HUM... I THINK HE'S IN BAD SHAPE !":"ANYWAY HE'S DOING OK !":
                        (status.current-1)?"ANYWAY HE'S LOOKINK GOOD !":
                        "I FEEL HE'S UP SHIT CREEK...",COUL_TEXT);
                        flag_dialogue = FALSE;
                      }
                  }
                else pass(); // l-othello passes his turn
              else // jeu fini
                { status.pause = TRUE;
                  couleur_courante(status.current);
                  setfillstyle(SOLID_FILL,COUL_RECT);
                  bar(282,32,598,58); // clear text zone
                  if (j1>j2)
                    { textxy(440,39,"\2 BLACK PLAYER WINS \2",COUL_TEXT);
                      textxy(440,51,j2?"HE'S A GIFTED PLAYER":"NO WHITE PIECES LEFT, EXCELLENT !!",COUL_TEXT);
                    }
                  else if (j1<j2)
                         { textxy(440,39,"\1 WHITE PLAYER WINS \1",COUL_TEXT);
                           textxy(440,51,j1?"IT WAS A NICE GAME":"NO MORE BLACK PIECES, WELL DONE !!",COUL_TEXT);
                         }
                       else
                         { textxy(440,39,"SORRY... IT'S A DRAW GAME.",COUL_TEXT);
                           textxy(440,51,"THEREFORE, NO WINNER !",COUL_TEXT);
                         }
                }
       x = 0;
       // if (kbhit()) ch = getch(); else ch  = FALSE;
     fflush(stdin);
     } while (!status.fin /* && ch != 27 */);

  closegraph();

  system("mode co80");  // to get back a normal 80 col screen

// Hey, don't panic. It's only an Ascii-art... see news: alt.ascii-art
   puts("        \\\\             /  Shareware");puts("         \\\\\\\' ,      / //");puts("          \\\\\\\//    _/ //'");
   puts("           \\\_-//' /  //<'             o-==- Lothello  v1.1Us -==o");
   puts("             \\ ///  <//'              e-mail: matthieu@hrnet.fr");puts(
   "              /  >>   \\\\\\\`            (c) By A.Rassat, M.Klein 1996");puts
   ("           /,)-^>>  _\\`");puts("           (/   \\\\ / \\\\\\\ ");
   puts("                 //  //\\\\\\\ ");puts("                ((`");puts("");
   puts("ThAnX tO: Stephane Breton, Fr (C Language teacher)");
   puts("          Christophe Corne, Fr (Director of project)");
   puts("          Edgar Swank, Us (Ideas, hi-test and americanization)");

  // at this line, the prog is terminated, and quits!
}

/* void interrupt display_node(void)
{  char j2c[20];

   disable();
   ltoa(count,j2c,10);
   setfillstyle(SOLID_FILL,BLACK);
   setcolor(LIGHTBLUE);
   if (status.current==1)    // BLACK player zone
    // affiche  compteur
     { bar(130,397,196,428);
       outtextxy(168,420,j2c); // Darn function !!
                               // the calling fucked up the Pc.
     }
   else                      // WHITE player zone
     { bar(197,397,275,428);
       outtextxy(229,420,j2c); // !! buggy here
     }
   enable();
   oldhandler(); // call interrupt. 0x1C (continue)

} */



/********************************** pass ************************************/

void pass()

{ octet possibles[33];

  bouton(377,430,1,"PASS");

  if (coup_possibles(damier,possibles,status.current))
    { setfillstyle(SOLID_FILL,COUL_RECT);
      bar(282,32,598,58);
      textxy(440,39,"YOU CAN USE THIS ONLY IF",COUL_TEXT);
      textxy(440,51,"YOU HAVE NO LEGAL MOVES",COUL_TEXT);
      sound(500); delay(65); sound(1000); delay(65); nosound();
    }
  else
    { sound(1000); delay(65);
      sound(500); delay(65);
      nosound(); delay(150);
      couleur_courante(Reverse(status.current));
    }

  bouton(377,430,0,"PASS");

}

/****************************************************************************
 *** This procedure applies a move on the board which address has been    ***
 *** transmitted to new_board. It is used when the player tries to place  ***
 *** a piece, and when the computer looks for moves recursively.          ***
 ****************************************************************************/

void apply_move(octet new_board[10][10], octet* move, int mode)

// new_board[][]  the board on which the move is applied
// move           <x, y, player>
// mode           TRUE = record the move in historique[]

// Notice that octet* move is given as a parameter, this just in order not
// to have a variable to redefine, which allows us to spare time, the
// procedure being called at each alpha-beta algo's node

{ int i,
      index_x, x,
      index_y, y,
      fin, ligne_entourable, nbr_pions_entoures;

  count++;

  for (i=0; i<=14; i+=2)  // one determines one of the 8 directions to scan
    { index_x = perimetre[i];
      index_y = perimetre[i+1];
      fin = ligne_entourable = nbr_pions_entoures = FALSE;
      x = Mask_x(*move); y = Mask_y(*move);
      do { x += index_x; y += index_y;
           if (!new_board[x][y]) fin = TRUE; // there is nothing in that
                                             // direction, or we reached
                                             // the border line                                        // a atteint la ceinture
           else if (new_board[x][y]==Mask_c(*move)) ligne_entourable = TRUE;
                else nbr_pions_entoures++;
         } while ( !fin && !ligne_entourable );

      if (ligne_entourable && nbr_pions_entoures)
        do { x -= index_x; y -= index_y;
             if (mode && (new_board[x][y]!=Mask_c(*move)))
               apply_pion(x,y,Mask_c(*move),1); // mode lent
             new_board[x][y] = Mask_c(*move);
           } while (nbr_pions_entoures--);
    }

  // synchronisation of the game's historical record with the move which has
  // been played.

  if (mode)
    if (*historique==historique[1])
      { historique[*historique+2] = *move;
        historique[1] = ++*historique;
        setfillstyle(SOLID_FILL,BLACK);
        bar(270,64,610,75); // erase "Use Forward/Backward to replay or continue"
      }
    else if (*move!=historique[historique[1]+2])
           { historique[historique[1]+2] = *move;
             *historique = ++historique[1];
             setfillstyle(SOLID_FILL,BLACK);
             bar(270,64,610,75); // erase "Use Forward/Backward to replay or continue"
           }
         else historique[1]++;

}


/******************************************************************
 *    Returns the possible moves' list at any time in the game    *
 *                       for a given player                       *
 ******************************************************************/

int coup_possibles(octet board[10][10], octet possibles[33], octet couleur)

{ int min_x = 8, min_y = 8, max_x = 1, max_y = 1,
      direct, index_x, x, index_y, y,
      fin, ligne_entourable, nbr_pions_entoures;

  // we initialize the prolog of the table POSSIBLES[] at 0
  octet i, j, nbr_coups_possibles = *possibles = 0;

// The following block determines a maximal square on the board in which
// we will look for the possible moves. We know for exmaples that if the
// 4 starting pieces are in the board's middle, it is not necessary to
// place pieces at the board's perimeter! Measures show that at the game's
// beginning, 'coup_possibles' gets 5 times faster with that block, and
// only gets 1.1 times slower at the end of the game.

/**************/
  for (j=1; j<=8; j++)
    for (i=1; i<=8; i++)
      if (board[i][j] == Reverse(couleur)) // Inverses the color given
        { if (min_x > i) min_x = i;        // as parameter, koz the maximal
          if (max_x < i) max_x = i;        // square is determined by the
          if (min_y > j) min_y = j;        // opposing pieces' disposition
          if (max_y < j) max_y = j;
        }

  if (min_x != 1) min_x--;
  if (max_x != 8) max_x++;
  if (min_y != 1) min_y--;
  if (max_y != 8) max_y++;
/**************/

  for (j=min_y; j<=max_y; j++)
    for (i=min_x; i<=max_x; i++)
        if (!board[i][j])
          for (direct=0; direct<=14; direct+=2)  // we determine one of the 8
                                                   // directions to scan.
            { index_x = perimetre[direct];
              index_y = perimetre[direct+1];
              fin = ligne_entourable = nbr_pions_entoures = FALSE;
              x = i; y = j;
              do { x += index_x; y += index_y;
                   if (!board[x][y]) fin = TRUE; // there is nothing in that
                                                 // direction, or we reached
                                                 // the border line
                   else if (board[x][y]==couleur) ligne_entourable = TRUE;
                        else nbr_pions_entoures++;
                 } while ( !fin && !ligne_entourable );

              if (ligne_entourable && nbr_pions_entoures)
                { direct = 14;
                  possibles[0] = ++nbr_coups_possibles;
                  possibles[possibles[0]] = ((i-1)<<5)|((j-1)<<2)|couleur;
                }
              }
  return(nbr_coups_possibles);
}


/************************** evaluation_damier ******************************/

int evaluation_damier(octet le_damier[][10])

// Macros for board evaluation procedure
#define Coin 500
#define if_case(x,y,n) if(le_damier[x][y])if(le_damier[x][y]==1)note+=n;else note-=n;
#define if_case_temp(x,y,n) if(temp[x][y])if(temp[x][y]==1)note+=n;else note-=n;
// #define Max_note 20000     // see at the begining of the prg
#define Coeff_mobilite 20
#define Coeff_degradation 1
#define Case_x (-200)
#define Case_c (-40)
#define Bord 75

{ register int note=0;

  octet temp[10][10], coups_j1[33], coups_j2[33], pions=0, i, j;

  char majoritaire = 0;

// Is the game finished ? (both players cannot play anymore). If so,
// and if BLACK won, then note = Max_note, else note = -Max_note
// If same number of both pieces, note = 0.

  coup_possibles(le_damier ,coups_j1, 1);
  coup_possibles(le_damier, coups_j2, 2);

  if (!*coups_j1 && !*coups_j2)
    { for (i=1; i<=8; i++)
        for (j=1; j<=8; j++)
          if (le_damier[i][j])
            if (le_damier[i][j] == 1)
              majoritaire++;
            else
              majoritaire--;
      if (majoritaire < 0)
        return(-Max_note);
      else if (majoritaire > 0)
             return(Max_note);
           else return (0);
    }

// Number of pieces counter. Let's assume that it is up to the computer
// to play. It may take a corner either immediately or later, because
// the human player cannot do anything against it in the x moves of the
// studied depth. The computer has therefore no reason to take it
// immediately. But the corner's value will decrease as fast as the number
// of pieces on the board increases. This therefore leads to a a faster take.

  for (i=1; i<=8; i++)
    for (j=1; j<=8; j++)
       if (le_damier[i][j] !=0) pions++;

// Stable pieces detection (edges only)

  memset(temp,0,100);

  if (le_damier[1][1])
    { i=2;
      temp[1][1] = le_damier[1][1];
      while(le_damier[i][1]) i++;
      if (i==9) for(i--;i>=2;i--) temp[i][1] = le_damier[i][1];
      else { i=2;
             while(le_damier[i][1] == le_damier[1][1])
              { temp[i][1] = le_damier[1][1]; i++; }
           }
      i=2;
      while(le_damier[1][i]) i++;
      if (i==9) for(i--;i>=2;i--) temp[1][i] = le_damier[1][i];
      else { i=2;
             while(le_damier[1][i] == le_damier[1][1])
              { temp[1][i] = le_damier[1][1]; i++; }
           }
    }

  if (le_damier[1][8])
    { i=2;
      while(le_damier[i][8]) i++;
      if (i==9) for(i--;i>=2;i--) temp[i][8] = le_damier[i][8];
      else { i=2;
             while(le_damier[i][8] == le_damier[1][8])
              { temp[i][8] = le_damier[1][8]; i++; }
           }
      if (!temp[1][8])  // if false, this edge is already full!
        { temp[1][8] = le_damier[1][8];
          i=7;
          while(le_damier[1][i] == le_damier[1][8])
               { temp[1][i] = le_damier[1][8]; i--; }
        }
    }

  if (le_damier[8][8])
    { if (!temp[8][8])
        { i=7;
          temp[8][8] = le_damier[8][8];
          while(le_damier[i][8] == le_damier[8][8])
           { temp[i][8] = le_damier[8][8]; i--; }
        }
      i=7;
      while(le_damier[8][i]) i--;
      if (i==0) for(i++;i<=7;i++) temp[8][i] = le_damier[8][i];
      else { i=7;
             while(le_damier[8][i] == le_damier[8][8])
              { temp[8][i] = le_damier[8][8]; i--; }
           }
    }

  if (le_damier[8][1])
    { i=8;
      while(le_damier[i][1] == le_damier[8][1])
         { temp[i][1] = le_damier[8][1]; i--;
         }
      i=2;
      while(le_damier[8][i] == le_damier[8][1])
         { temp[8][i] = le_damier[8][1]; i++;
         }
    }

 // we now establish the points to give to the temp[][] stable pieces

  for (i=1;i<=8;i++)
    { if_case_temp(i,1,Bord)
      if_case_temp(i,8,Bord)
    }
  for (i=2;i<=7;i++)
    { if_case_temp(1,i,Bord)
      if_case_temp(8,i,Bord)
    }

 // supplement for the corners. A corner's value must be greater than
 // the sum of the 6 pieces of a board (w/o the corner)...

  if_case_temp(1,1,Coin-Coeff_degradation*pions)
  if_case_temp(1,8,Coin-Coeff_degradation*pions)
  if_case_temp(8,1,Coin-Coeff_degradation*pions)
  if_case_temp(8,8,Coin-Coeff_degradation*pions)


 // X and C squares scanning...a few negative points are given.
 // We could have written tests which work with symmetry...

  if (!le_damier[1][1])  { if_case(2,2,Case_x)
                           if_case(1,2,Case_c)
                           if_case(2,1,Case_c)
                         }

  if (!le_damier[8][8])  { if_case(7,7,Case_x)
                           if_case(8,7,Case_c)
                           if_case(7,8,Case_c)
                         }

  if (!le_damier[1][8])  { if_case(2,7,Case_x)
                           if_case(1,7,Case_c)
                           if_case(2,8,Case_c)
                         }

  if (!le_damier[8][1])  { if_case(7,2,Case_x)
                           if_case(7,1,Case_c)
                           if_case(8,2,Case_c)
                         }

  // Relative mobility rating

  note += (Coeff_mobilite*(*coups_j1-*coups_j2));
  return(note);
}


/********************************* alpha_beta ******************************/

int alpha_beta(octet board[][10], octet depth, octet joueur, octet* move, int* score, int note, octet who)

// who is used to know wether 2 following tree generations are Max then Min
// or Min then Max, because it may happen that we have Min-Min or Max-Max,
// which are cases where alpha-beta cannot be applied.

{ int i, the_score;
  octet new_board[10][10], liste[33], the_move;

  if (mouse_bout()) return (FALSE);

  if (!coup_possibles(board,liste,joueur))
   if (!coup_possibles(board, liste, joueur = Reverse(joueur)))
     { *score = evaluation_damier(board); return(TRUE);
     }

  if (joueur == 1) *score = -Max_note; else *score = Max_note;

  depth--;

  for (i=1; i<=*liste; i++)
    { memcpy(new_board, board, 100);
      apply_move(new_board, liste+i, FALSE);
      if (depth) { if (!alpha_beta(new_board,
                                   depth,
                                   Reverse(joueur),
                                   &the_move,
                                   &the_score,
                                   *score,
                                   joueur))
                     return(FALSE);
                 }
      else the_score = evaluation_damier(new_board);

      if (joueur == 1)
        { if (who == 2 && the_score >= note)
            { // min->max cut-off
              i=*liste; *score = Max_note;
            }
          else if (the_score >= *score)
                 { *score = the_score;
                   *move = liste[i];
                  if (*score == Max_note) i=*liste; // end of game cut-off
                 }
        }
      else
        { if (who == 1  && the_score <= note)
            { // max->min cut-off
              i=*liste; *score = -Max_note;
            }
          else if (the_score <= *score)
                 { *score = the_score;
                   *move = liste[i];
                   if (*score == -Max_note) i=*liste; // end of game cut-off
                 }
        }
    }

  return(TRUE);     // phewww ..!.

}

/********************************* minimax **********************************/

int minimax(octet board[][10], octet depth, octet joueur, octet* move, int* score)

{ int i, the_score;
  octet new_board[10][10], liste[33], the_move;

  if (mouse_bout()) return (FALSE);

  if ( !coup_possibles(board,liste,joueur) )
   if ( !coup_possibles(board, liste, joueur = Reverse(joueur)) )
       { *score = evaluation_damier(board); return(TRUE);
       }

  if (joueur == 1) *score = -Max_note; else *score = Max_note;

  depth--;

  for (i=1; i<=*liste; i++)
    { memcpy(new_board, board, 100);
      apply_move(new_board, liste+i, FALSE);
      if (depth)
        { if (!minimax(new_board,
                       depth,
                       Reverse(joueur),
                       &the_move,
                       &the_score))
            return(FALSE);
        }
      else the_score = evaluation_damier(new_board);

      if (joueur == 1)
        { if (the_score >= *score)
            { *score = the_score;
              *move = liste[i];
              if (*score == Max_note) i=*liste; // end of game cut-off
            }
        }
      else if (the_score <= *score)
             { *score = the_score;
               *move = liste[i];
               if (*score == -Max_note) i=*liste; // end of game cut-off
             }
    }

  return(TRUE);

}

/**************************** presentation **********************************/

void presentation()

{ int gdriver = DETECT,
      gmode,
      i,j,k,
      poly_d[8] = { 280, 80, 600, 80, 600, 400, 280, 400 },
      poly_g[8] = { 29, 80, 211, 80, 211, 320, 29, 320 },
      x = 570, y = 419; // pour les boutons fleches

  char hint[] = "HINT(xx)";

  clrscr();

  mouse_curs(0); // init mouse and return register values
  if (!Regs.x.ax)
    { puts("Hey ! Mouse driver not detected..."); exit(1);
    }
  else  // Mouse info...(Book reference: 'Bible PC', by Michael Tischer)
    { textattr(79);
      cprintf("                 Little Othello: Hell on Checkerboard v1.1Us                    ");
      Regs.x.ax=0x24;int86(0x33,&Regs,&Regs);printf("\nMouse driver v %u.%u detected"
      ,Regs.h.bh,Regs.h.bl);gmode=Regs.h.ch;printf("\nType  : %s",gmode==1?
      "Bus":gmode==2?"Serial":gmode==3?"InPort":gmode==4?"PS/2":
      "Hewlett-Packard");printf("\nIrq   : %u",Regs.h.cl); delay(100);
      // just for fun: a 'DooM like' init
      puts("\nM_Init: Init miscellaneous info.");
      puts("R_Init: Init LOTH refresh daemon - [                                      ]");
      gotoxy(37,7); i=38;
      while(i--) { printf("."); delay(70); }
      puts("\nP_Init: Init Playloop state."); delay(500);
      // next is a subliminal message, for psychological preparation... :)
      puts("You can Win..."); delay(25);
    }

  if (registerbgidriver(EGAVGA_driver)<0) exit (1);
  if (registerbgifont(triplex_font)<0) exit (1);
  initgraph(&gdriver,&gmode,"");  // init screen
  if (graphresult() != grOk ) exit(1);

//// 1st page intro
  settextjustify(CENTER_TEXT,CENTER_TEXT);
  setcolor(RED);
  setbkcolor(BLACK);
  rectangle(0,0,getmaxx(),getmaxy());
  rectangle(1,1,getmaxx()-1,getmaxy()-1);
  textxy(436,464,"UHA - FST - Deug ][ Technologie Industrielle",WHITE);
  textxy(436,449,"Lothello v1.1Us by Matthieu KLEIN & Andre RASSAT",WHITE);
  textxy(350,190,"TM",WHITE);
  settextstyle(TRIPLEX_FONT,HORIZ_DIR,7);
  textxy(getmaxx()/2,getmaxy()/2-40,"OTHELLO GAME",YELLOW);
  textxy(getmaxx()/2-1,getmaxy()/2-41,"OTHELLO GAME",LINE2);
  Erase_KeybBuff
  while (!kbhit()&&!mouse_bout()); // press a key or mouse button to quit

//// now, draw board and init 1st value
  cleardevice();   // erase graphic screen
  mouse_win(0,getmaxx()-10,0,getmaxy()-20);  // set mouse window

//// arrow button base
loop:
  setfillstyle(SOLID_FILL,LIGHTGRAY);
  bar(x+2,y+2,x+16,y+16);
  setfillstyle(SOLID_FILL,DARKGRAY);
  bar(x+3,y+3,x+15,y+15);
  setfillstyle(SOLID_FILL,LIGHTGRAY);
  bar(x+4,y+4,x+14,y+14);
  setcolor(WHITE);
  line(x+3,y+3,x+14,y+3); line(x+3,y+4,x+3,y+14);
  setcolor(BLACK);
  if (y==419)
    { moveto(x+5,y+9); // perimeter black up arrow
      lineto(x+8,y+9); lineto(x+8,y+13); lineto(x+10,y+13); lineto(x+10,y+9);
      lineto(x+13,y+9); lineto(x+9,y+5); lineto(x+5,y+9);
      setcolor(LIGHTRED);
      line(x+9,y+12,x+9,y+6); line(x+7,y+8,x+11,y+8); line(x+8,y+7,x+10,y+7);
      y = 444; goto loop;
    }
  else
    { moveto(x+8,y+5); // perimeter black down arrow
      lineto(x+8,y+9); lineto(x+5,y+9); lineto(x+9,y+13); lineto(x+13,y+9);
      lineto(x+10,y+9); lineto(x+10,y+5); lineto(x+8,y+5);
      setcolor(LIGHTRED);
      line(x+9,y+6,x+9,y+12); line(x+7,y+10,x+11,y+10); line(x+8,y+11,x+10,y+11);
    }
  fleches(0,0); fleches(1,0);

  setcolor(COUL_RECT);
  setfillstyle(SOLID_FILL,COUL_RECT);
  fillpoly(4,poly_d);   // draw and fill a polygon (for the board)
  setfillstyle(SOLID_FILL,COUL_RECT);
  fillpoly(4,poly_g);   // draw and fill a polygon (for the left buttons)
  bar(280,30,600,60);   // fill a rectangle above the board
  bar(280,420,551,460); // fill a rectangle below the board (3 buttons)

  //// shade for rectangles + board + board square pattern
  setcolor(LINE1);
  rectangle(30,81,212,321);   // shade left zone buttons
  rectangle(281,31,601,61);   // shade text zone above board
  rectangle(281,421,552,461); // shade buttons zone below board
  for (i=0, j=0; i<9; i++, j+=40)
    { line(poly_d[0]+1,poly_d[1]+1+j,poly_d[2]+1,poly_d[3]+1+j);
      line(poly_d[0]+1+j,poly_d[1]+1,poly_d[6]+1+j,poly_d[7]+1);
    }  // draw board perimeter + first square pattern

  //// rectangles perimeter + board +  board square pattern
  setcolor(LINE2);
  rectangle(28,79,210,319); rectangle(29,80,211,320); // left zone buttons
  rectangle(279,29,599,59); rectangle(280,30,600,60); // text zone above board
  rectangle(279,419,550,459); rectangle(280,420,551,460); // buttons zone below board

  for (i=0,k=0;i<9;i++,k+=40)
    for (j=-1;j<1;j++)
      { line(poly_d[0]+j,poly_d[1]+k+j,poly_d[2]+j,poly_d[3]+k+j);
        line(poly_d[0]+k+j,poly_d[1]+j,poly_d[6]+k+j,poly_d[7]+j);
      }

  //// display the 2 pieces in top buttons zone
  setcolor(WHITE);                   // white perimeter for black piece
  setfillstyle(SOLID_FILL,BLACK);
  fillellipse(80,110,19,19);         // draw the black piece
  setcolor(BLACK);                   // black perimeter for white piece
  setfillstyle(SOLID_FILL,WHITE);
  fillellipse(160,110,19,19);        // draw the white piece

  //// top left 'Othello' display
  settextstyle(TRIPLEX_FONT,HORIZ_DIR,7);
  textxy(120,32,"Othello",LIGHTCYAN);
  textxy(119,31,"Othello",LINE2);
  settextstyle(DEFAULT_FONT,HORIZ_DIR,1);
  textxy(220,20,"TM",WHITE);

  //// status var. init
  status.fin = FALSE;
  status.joueur[0].humain = 1;  // joueur 1 = human
  status.joueur[1].humain = 0;
  status.joueur[1].niveau = 2;  // (1=novice, 2=meduim, 3=expert, 4=user)
  status.hint_level = 3;
  status.joueur[0].user = 2;
  status.joueur[1].user = 6;
  status.algo = FALSE; // alpha-beta / min-max (press CTRL to toggle !)
  historique[62] = 2;
  refresh_buttons();

  //// '-' & '+' for the USER depth control
  textxy(35,298,"+",WHITE);  textxy(35,299,"+",BLACK);
  textxy(35,308,"-",WHITE);  textxy(35,309,"-",BLACK);
  textxy(203,298,"+",WHITE);  textxy(203,299,"+",BLACK);
  textxy(203,308,"-",WHITE);  textxy(203,309,"-",BLACK);

  setcolor(WHITE);
  line(42,199,198,199);
  setcolor(BLACK);
  line(42,200,198,200);

  //// draw buttons
  bouton(377,430,0,"PASS");
  memcpy(hint+5+strlen(itoa(status.hint_level,hint+5,10)),")\0",2);
  bouton(462,430,0,hint); // 'HINT' button
  textxy(543,446,"-",WHITE);  textxy(543,447,"-",BLACK);
  textxy(543,436,"+",WHITE);  textxy(543,437,"+",BLACK);
  bouton(12,440,0,"SAVE"); bouton(92,440,0,"LOAD");
  bouton(172,440,0,"QUIT");

  //// for infos about mobility and nbr of pieces
  setcolor(LINE2);
  settextjustify(LEFT_TEXT,CENTER_TEXT);
  outtextxy(150,345,"BLACK"); outtextxy(210,345,"WHITE");
  outtextxy(10,360,"Piece(s)      >");
  outtextxy(10,375,"Mobility      >");
  outtextxy(10,390,"Depth (Pc)    >");
  outtextxy(10,405,"Board eval.   >");
  outtextxy(10,420,"Node(s) tree  >");
  textxy(280,411,"Alpha-Beta enabled (Ctrl to swap)",WHITE);
  settextjustify(CENTER_TEXT,CENTER_TEXT);

  //// set the mouse
  mouse_set(240,220);       // we set the mouse at this position
  mouse_curs(1);            // cursor display

  init_game();

}

/******************************** init_game *********************************/

void init_game()

{ setfillstyle(SOLID_FILL,BLACK);
  bar(270,64,610,75); // erase "Use Forward/Backward to replay or continue"

  bouton(291,430,1,"NEW GAME");
  memset(damier,0,100);
  status.pause = FALSE;

//// initialize the game's historical record
  *historique = historique[1] = NULL;

//// thanks to historique[62], we alternate the 4 starting pieces, once out of two
  historique[62] = Reverse(historique[62]);
  damier[4][4] = damier[5][5] = Reverse(historique[62]);
  damier[4][5] = damier[5][4] = historique[62];
  dessine_matrice_damier();        // and we draw

//// message info
  setfillstyle(SOLID_FILL,COUL_RECT);
  bar(282,32,598,58); // erase old text
  textxy(440,39,"CLICK PIECE OF THE COLOR THAT BEGINS",COUL_TEXT);
  textxy(440,51,"OR PRESS NEW GAME BUTTON FOR INIT GAME",COUL_TEXT);

//// erase red circle
  setcolor(COUL_RECT);
  circle(160,110,23); circle(160,110,24);
  circle(80,110,23); circle(80,110,24);

  couleur_courante(0);
  bouton(291,430,0,"NEW GAME");
}


/********************************* fleches **********************************/

void fleches(int bouton, int etat)

{ int x=570, y=419+bouton*25;

// draw the arrow button's perimeter
  mouse_curs(2); // hide mouse
  setcolor(etat*WHITE+DARKGRAY*!etat);
  line(x+1,y+17,x+17,y+17); line(x+17,y+16,x+17,y+1);
  setcolor(WHITE*!etat+etat*DARKGRAY);
  line(x+16,y+1,x+1,y+1); line(x+1,y+2,x+1,y+16);
  mouse_curs(1); // show mouse

}

/**************************** couleur_courante ******************************/

void couleur_courante(int color)

{ mouse_curs(2); // hide mouse
  flag_info = flag_dialogue = TRUE; // enable the 'who has won' dialog
  setfillstyle(SOLID_FILL,status.pause?LIGHTRED:COUL_RECT);

  switch (color)
   { case 1 : setcolor(status.pause?BLACK:LIGHTRED); circle(80,110,23); circle(80,110,24);
              setcolor(COUL_RECT); circle(160,110,23); circle(160,110,24);
              bar(282,32,598,58); // clear text zone
              textxy(440,39,status.pause?"PAUSE: BLACK PIECES WAITING... PRESS":"BLACK PIECES",COUL_TEXT);
              textxy(440,51,status.pause?"RIGHT MOUSE BUTTON TO START/PAUSE":"ARE PLAYING NOW",COUL_TEXT);
              status.current = color; break;
     case 2 : setcolor(status.pause?BLACK:LIGHTRED); circle(160,110,23); circle(160,110,24);
              setcolor(COUL_RECT); circle(80,110,23); circle(80,110,24);
              bar(282,32,598,58); // clear text zone
              textxy(440,39,status.pause?"PAUSE: WHITE PIECES WAITING... PRESS":"WHITE PIECES",COUL_TEXT);
              textxy(440,51,status.pause?"RIGHT MOUSE BUTTON TO START/PAUSE":"ARE PLAYING NOW",COUL_TEXT);
              status.current = color; break;
     default: status.current = 0;
   }

  mouse_curs(1);

}

/****************************** pixel2case **********************************/

void pixel2case(int pixel_x, int pixel_y, int* coup_x, int* coup_y)

{ int out=TRUE, i, j, cases[8] = { 282, 82, 318, 82, 318, 118, 282, 118 };

  for (j=0,*coup_y=0; j<320; j+=40,++*coup_y)
    for (i=0,*coup_x=0; i<320; i+=40,++*coup_x)
      if ((pixel_x>cases[0]+i)&&(pixel_x<cases[2]+i)&&(pixel_y>cases[1]+j)&&(pixel_y<cases[7]+j))
          { out=FALSE; i=j=321;
          }
      else if ((pixel_x<cases[0]+i)&&(pixel_y<cases[1]+j))
             { i=j=321;
             }
  if (out) *coup_x = NULL; // if *coup_x = 0 the square is not valid
}

/************************ dessine_matrice_damier ****************************/

void dessine_matrice_damier()

{ int i, j=1;

//// draw on screen the board corresponding to the matrix

  for(; j<=8; j++)
    for(i=1; i<=8; i++)
       apply_pion(i, j, damier[i][j], 0); // fast mode

}

/****************************** apply_pion **********************************/

void apply_pion(int x, int y, int couleur, int mode)

{ int i,pixel_x,pixel_y,offset=FALSE;

//// Applies a piece on the graphical board at the x,y coordinates
  // mode = 0 fast display
  //        1 slow display
  //        2 display dots instead of piece
  //        3 erase dots

  if (mode == 0 || mode == 1)
    { mouse_pos(&pixel_x,&pixel_y);
      if ((pixel_x>281)&&(pixel_x<598)&&(pixel_y>81)&&(pixel_y<399))
        { mouse_curs(2);  // hide mouse
          offset=TRUE;
        }
      if (couleur)
        { setcolor((couleur==1)?WHITE:BLACK);
          setfillstyle(SOLID_FILL,(couleur==1)?BLACK:WHITE);

          if(mode) // if mode == TRUE then slow piece display, else fast
            for(i=0;i<=18;i++)
              { fillellipse(40*x+260,40*y+60,i,i);  // we lay the piece down on the board
                // delay(20);
              }
          else  fillellipse(40*x+260,40*y+60,18,18);
        }
      else
        { setfillstyle(SOLID_FILL,COUL_RECT);
          bar(40*x+242,40*y+42,40*x+278,40*y+78);
        }
    }
  else { if (mode==2)
            { setcolor(BLACK);
              setfillstyle(SOLID_FILL,couleur);
              fillellipse(40*x+260,40*y+60,4,4);
            }
         else { setfillstyle(SOLID_FILL,COUL_RECT);
                bar(40*x+256,40*y+56,40*x+264,40*y+64);
              }
       }

  if (offset) mouse_curs(1); // show mouse

}

/********************************** textxy **********************************/

void textxy(int x, int y, char* chaine, int couleur)

{ setcolor(couleur);
  outtextxy(x,y,chaine);
}

/*********************************** bouton *********************************/

void bouton(int x, int y, int on_off, char* caption)

{ int i=0, tab[8] = {0, 0, 76, 0, 76, 20, 0, 20 };

  for (;i<=7;i+=2)
    { tab[i]+=x;
      tab[i+1]+=y;
    }

  mouse_curs(2);           // hide mouse cursor.
  setcolor(BLACK);
  setfillstyle(SOLID_FILL,LIGHTGRAY);
  fillpoly(4,tab);
  setcolor(on_off?DARKGRAY:WHITE);
  line(tab[0],tab[1],tab[2],tab[3]);
  line(tab[0],tab[1],tab[6],tab[7]-1);
  setcolor(on_off?WHITE:DARKGRAY);
  line(tab[6]+1,tab[7]-1,tab[4],tab[5]-1);
  line(tab[4],tab[5]-1,tab[2],tab[3]+1);
  textxy((2*x+76)/2,(2*y+20)/2,caption,on_off?BLACK:WHITE);
  textxy((2*x+76)/2,(2*y+20)/2+1,caption,on_off?WHITE:BLACK);
  mouse_curs(1);          // show mouse cusor
}

/********************************** gest_sour *******************************/

void gest_sour(int *damier_x, int *damier_y)

{ int old_damier_x = 0,
      old_damier_y = 0,
      x, y, i,
      aide = FALSE,
      bout_fleche = TRUE;
  octet possible[33];
  char user[] = "USER(xx)";
  //// !! writing 'char* user = "USER(xx)"' makes Turbo C++ 1.01 bug
  //// with 'null pointer assignement' ?

  if (mouse_bout()==1)   // button 1 was clicked
    { mouse_curs(2); // hide mouse
      for(x=1; x<=2; x++)
        { coup_possibles(damier,possible,x);
          for(i=1;i<=*possible;i++) // erase the opposing possible moves dots
            apply_pion(Mask_x(possible[i]),Mask_y(possible[i]),0,3);
        }
      mouse_curs(1); // show mouse

      do { mouse_pos(&x,&y);

           if (x>570 && x<588 && bout_fleche)  // arrow buttons zone
             if (y>419 && y<437) // up arrow button (Backwards)
               { bout_fleche = FALSE;
                 if (historique[1])
                   { fleches(0,1); historique[1]--; memset(damier,0,100);
                     damier[4][4] = damier[5][5] = Reverse(historique[62]);
                     damier[4][5] = damier[5][4] = historique[62];
                     for (i=2; i<=historique[1]+1; i++)
                       apply_move(damier, historique+i, 0);
                     dessine_matrice_damier();
                     status.pause = TRUE; // Pause ON
                     couleur_courante(Mask_c(historique[historique[1]+2]));
                   }
               }
             else if (y>444 && y<462)   // down arrow button (Forwards)
                   { bout_fleche = FALSE;
                     if (*historique != historique[1])
                       { fleches(1,1); historique[1]++; memset(damier,0,100);
                         damier[4][4] = damier[5][5] = Reverse(historique[62]);
                         damier[4][5] = damier[5][4] = historique[62];
                         for (i=2; i<=historique[1]+1; i++)
                           apply_move(damier, historique+i, 0);
                         dessine_matrice_damier();
                         status.pause = TRUE; // Pause ON
                         if (*historique==historique[1])
                           { couleur_courante(Reverse(Mask_c(historique[historique[1]+1])));
                             if (!coup_possibles(damier,possible,1)&&!coup_possibles(damier,possible,2))
                               status.pause = FALSE;
                           }
                         else
                           couleur_courante(Mask_c(historique[historique[1]+2]));
                       }
                   }

               if (x>281 && x<598 && y>81 && y<399) // && status.joueur[status.current-1].humain && status.current)
               { pixel2case(x, y, damier_x, damier_y);
                 if (damier[*damier_x][*damier_y])
                  { coup_possibles(damier,possible,damier[*damier_x][*damier_y]);
                    for(i=1;i<=*possible;i++) // display the possible moves dots
                     apply_pion(Mask_x(possible[i]),Mask_y(possible[i]),BLACK,2);

                    // we use an intermediate var. old_damier to allow
                    // the player to slide the mouse on the opposing pieces,
                    // by erasing the dots for his own color.
                    // we cannot suppress the opponent's dots at each time
                    // since the common dots would sparkle.

                    if (old_damier_x!=*damier_x||old_damier_y!=*damier_y)
                     { old_damier_x = *damier_x; old_damier_y = *damier_y;
                       coup_possibles(damier,possible,Reverse(damier[*damier_x][*damier_y]));
                       for(i=1;i<=*possible;i++) // efface les points des cases possibles adverses
                         apply_pion(Mask_x(possible[i]),Mask_y(possible[i]),0,3);
                     }
                    aide = TRUE;
                  }
               }
             }
        while (mouse_bout());    // until mouse button released

        if (!bout_fleche) { fleches(0,0); fleches(1,0); } // arrow buttons OFF

        if (aide) // erase the possible moves dots
          { for(aide=1; aide<=8; aide++) // (var. 'aide' used for 'j')
              for(i=1; i<=8; i++)
                if (!damier[i][aide]) apply_pion(i, aide,0,3);
          }

        mouse_pos(&x,&y);     // we search (again) the mouse's position


        if (!status.current)
          if (y>90 && y<130)    // click on one of the 2 pieces above the
                                // ctrl buttons zone
            if (x>60 && x<100)  couleur_courante(1);
            else if (x>140 && x<180) couleur_courante(2);


        if (x>42 && x<118)        //  1st buttons column zone
          i = 0;
        else if (x>122 && x<198)  // 2nd buttons column zone
          i = 1;
        else i = 3;

        if (i==0 || i==1)
             {  if (y>144 && y<164) // Human button ON
                      { status.joueur[i].humain = 1;
                        refresh_buttons();
                      }
                else if (y>172 && y<192) // PC button ON
                      { status.joueur[i].humain = 0;  // Pc == ON
                        status.joueur[i].niveau = 2;  // default level
                        refresh_buttons();
                      }
                else if (y>208 && y<228) // Novice button ON
                      { status.joueur[i].niveau = 1;
                        status.joueur[i].humain = 0;
                        refresh_buttons();
                      }
                else if (y>236 && y<256) // Medium button ON
                      { status.joueur[i].niveau = 2;
                        status.joueur[i].humain = 0;
                        refresh_buttons();
                      }
                else if (y>264 && y<284) // Expert button ON
                      { status.joueur[i].niveau = 3;
                        status.joueur[i].humain = 0;
                        refresh_buttons();
                      }
                else if (y>292 && y<312) // User button ON
                      { status.joueur[i].niveau = 4;
                        status.joueur[i].humain = 0;
                        refresh_buttons();
                      }
             }

        if (y>292 && y<312)
          { if (x>29 && x<41)
              { if (y>302)
                  { if (status.joueur[0].user!=1) status.joueur[0].user--;
                  }
                else if (status.joueur[0].user!=60) status.joueur[0].user++;
                Waitforretrace();
                memcpy(user+5+strlen(itoa(status.joueur[0].user,user+5,10)),")\0",2);
                bouton(42,292,(status.joueur[0].niveau==4)?1:0,user);
              }
            else if (x>197 && x<209)
              { if (y>302)
                  { if (status.joueur[1].user!=1) status.joueur[1].user--;
                  }
                else if (status.joueur[1].user!=60) status.joueur[1].user++;
                Waitforretrace();
                memcpy(user+5+strlen(itoa(status.joueur[1].user,user+5,10)),")\0",2);
                bouton(122,292,(status.joueur[1].niveau==4)?1:0,user);
              }
            flag_info = TRUE;
          }

        if (y>440 && y<460) // SAVE / LOAD / QUIT
          if (x>12 && x<88) save_load(0);       // SAVE
          else if (x>92 && x<168) save_load(1); // LOAD
          else if (x>172 && x<248)              // QUIT
                 { bouton(172,440,1,"QUIT");
                   delay(60);
                   bouton(172,440,0,"QUIT");
                   status.fin = TRUE;// quit
                 }

        if (y>430 && y<450) // 3 buttons below the board zone
                {
                if (x>291 && x<366) // 'NEW GAME' button ON
                  init_game();
                else if ( x>376 && x<452 && status.current && !status.pause)
                     // Boutton Passe ON
                        pass();  // passes the turn (here with pre-ctrl)
                else if (x>462 && x<538)  // QUIT button ON
                        { if (status.joueur[status.current-1].humain)
                            if (coup_possibles(damier, possible, status.current))
                            { memcpy(user,"HINT",4);
                              memcpy(user+5+strlen(itoa(status.hint_level,user+5,10)),")\0",2);
                              bouton(462,430,1,user);
                              if (alpha_beta(damier,
                                             status.hint_level,
                                             status.current,
                                             possible,  // &coup
                                             &i,        // &note
                                             0, 0)
                                  )
                                 { mouse_curs(2); // hide mouse
                                   apply_pion(Mask_x(*possible),Mask_y(*possible),RED,2);
                                   mouse_curs(1); // show mouse
                                 }
                              bouton(462,430,0,user);

                            }


                        }
                else if (x>538 && x<549) // '+' & '-'
                       { Waitforretrace();
                         memcpy(user,"HINT",4);
                         if (y>440)
                           { if (status.hint_level!=1) status.hint_level--;
                           }
                         else
                           { if (status.hint_level!=60) status.hint_level++;
                           }
                         memcpy(user+5+strlen(itoa(status.hint_level,user+5,10)),")\0",2);
                         bouton(462,430,0,user);
                       }
                }
        else if (x>281 && x<598 && y>81 && y<399) // Board zone
               { pixel2case(x, y, damier_x, damier_y);
                 if (!status.current)  // used to determine which
                                       // color begins.
                  couleur_courante(damier[*damier_x][*damier_y]);
                }
        }

  if (mouse_bout()==2) // toggle pause
    { while(mouse_bout());
      if (status.current) status.pause = !status.pause;
      couleur_courante(status.current);
    }


  // Toggle alpha-beta <-> MinMax mode (CTRL key)
  // + secret message... (Shift key) Just for fun...
  if (Regs.h.ah = 0x02, int86(0x16, &Regs, &Regs), Regs.h.al & 4)
    { status.algo =! status.algo;
      settextjustify(LEFT_TEXT,CENTER_TEXT);
      setfillstyle(SOLID_FILL,BLACK);
      bar(280,402,640,416);
      if (status.algo) textxy(280,411,"Min-Max Search enabled (slow)",WHITE);
      else textxy(280,411,"Alpha-Beta Search enabled",WHITE);
      settextjustify(CENTER_TEXT,CENTER_TEXT);
      while (Regs.h.ah = 0x02, int86(0x16, &Regs, &Regs), Regs.h.al & 0xF);
    }
  else if (Regs.h.al & 3)
         { message();
           while (Regs.h.ah = 0x02, int86(0x16, &Regs, &Regs), Regs.h.al & 0xF);
         }
}

/****************************** message *************************************/

void message() // just for fun. (Press SHIFT Key)

{ /* 2 messages encrypted with XOr mask process.
     Will you be able to crack the code ? ;)
     The password is a french sentence, so is the result.
     There are 2 sentences, one in line 1, one in line 2,
     and one passphrase for each sentence.
     found ? send an e-mail to matthieu@hrnet.fr ! thAnXX :)
     */

  char s1[30]={38,15,9,85,16,69,5,73,47,0,22,23,69,30,27,
  85,23,84,76,4,83,110,10,10,1,71,77,68,82,0},s2[30]={55,
  7,2,0,44,28,26,27,85,22,84,45,8,26,0,16,88,83,20,82,77,
  13,65,99,28,13,8,21,0,0},pass[32]={30};setfillstyle(
  SOLID_FILL,COUL_RECT);bar(282,32,598,58);Erase_KeybBuff
  textxy(440,39,"Enter password to decode the message",
  COUL_TEXT);sound(1000);delay(100);nosound();cgets(pass);
  for(*pass=2;s1[-2+*pass]^=pass[*pass],s2[-2+*pass]^=pass
  [*pass],*pass<=30;++*pass);bar(282,32,598,58);textxy(440
  ,39,s1,COUL_TEXT);textxy(440,51,s2,COUL_TEXT);
}

/****************************** refresh_buttons *****************************/

void refresh_buttons()

{ int i;
  char user[] = "USER(xx)";

  for (i=0; i<=1; i++)
     { Waitforretrace();
       memcpy(user+5+strlen(itoa(status.joueur[i].user,user+5,10)),")\0",2);
       bouton(i?122:42,144,status.joueur[i].humain,"HUMAN");
       bouton(i?122:42,172,!status.joueur[i].humain,"PC");
       status.joueur[i].niveau *= !status.joueur[i].humain;
       bouton(i?122:42,208,status.joueur[i].niveau!=1?0:1,"NOVICE");
       bouton(i?122:42,236,status.joueur[i].niveau!=2?0:1,"MEDIUM");
       bouton(i?122:42,264,status.joueur[i].niveau!=3?0:1,"EXPERT");
       bouton(i?122:42,292,status.joueur[i].niveau!=4?0:1,user);
     }

  flag_info = TRUE;

}

/******************************** save_load *********************************/

void save_load(int mode)

{ // mode = 0: save
  // mode = 1: load

  char *info = "Filename:         .SAV  (Enter)",
       dos_filename[14] = "\0--------$$$$\0",
       ch = 0;
  unsigned char *lst;
  FILE* fp; // FilePointer
  void far *bitmap;
  struct ffblk ffblk;
  int done,
      i;

  Erase_KeybBuff

  bitmap = farmalloc(imagesize(282,32,598,58));
  getimage(282, 32, 598, 58, bitmap); // to save the current message

  setfillstyle(SOLID_FILL,LIGHTRED);

  if (mode)  // *** LOAD ***
    { bar(282,32,598,58); // efface zone texte
      bouton(92,440,1,"LOAD");
      lst = malloc(2);
      *lst = *(lst+1) = 0; // *lst = #files, *(lst+1) = internal list pointer
      done = findfirst("*.SAV",&ffblk,0);
      while (!done)
        { lst = realloc(lst,++*lst*13+2); // len('$$$$$$$$.SAV\0') = 13
          memcpy(lst+2+(*lst-1)*13,ffblk.ff_name,13);
          done = findnext(&ffblk);
        }
      textxy(440,39,"Select file with key cursor Up/Down",COUL_TEXT);
      setfillstyle(SOLID_FILL,LIGHTRED);
      if (*lst)
        { do { memcpy(info+10,"        ",8);
             i=0;
             while(*(lst+2+*(lst+1)*13+i)!='.')
             // right to left expression evaluation in memcpy() !!
             memcpy(info+10+i++,lst+2+*(lst+1)*13+i,1);
             bar(397,45,462,55); // erase '        '.SAV txt info
             textxy(440,51,info,COUL_TEXT);
             ch = getch();
             if (!ch) // then ch == 0
               { if ((ch=getch())==80) // down cursor key = next file
                   { if(*(lst+1)!=*lst-1) ++*(lst+1);
                   }
                 else if (ch==72) // key up = prev. file
                      if(*(lst+1)) --*(lst+1);
                 ch = 0;
               }
           } while (ch!=27 && ch!=13);
           if (ch==13)
             { memcpy((dos_filename),lst+2+*(lst+1)*13,13);
               fp=fopen(dos_filename,"rb");
               fread(historique,sizeof(historique),1,fp);
               fclose(fp);
               historique[1] = 0; memset(damier,0,100);
               damier[4][4] = damier[5][5] = Reverse(historique[62]);
               damier[4][5] = damier[5][4] = historique[62];
               dessine_matrice_damier();
               status.pause = TRUE; // Pause ON
               couleur_courante(Mask_c(historique[2]));
               textxy(440,70,"Use Forward/Backward to replay or continue",LIGHTBLUE);
             }
           else
             { mouse_curs(2);
               putimage(282, 32, bitmap, COPY_PUT);
               mouse_curs(1);
             }
         }
      else { textxy(440,51,"Sorry, no file found",WHITE);
             while(!kbhit()&&!mouse_bout()); while(mouse_bout());
             mouse_curs(2);
             putimage(282, 32, bitmap, COPY_PUT);
             mouse_curs(1);
           }
       }

  else if (*historique) // *** SAVE *** : no save if no game.
    { bar(282,32,598,58); // clear text zone
      bouton(12,440,1,"SAVE");
      textxy(440,39,"Enter Filename or ESC to quit",COUL_TEXT);
      setfillstyle(SOLID_FILL,WHITE); // for bar();
      bar(397,45,462,55); // erase '--------' with white box
      memcpy(info+10,dos_filename+1,8);
      textxy(441,51,info,COUL_TEXT);
      while (ch = getch(), ch!=13 && ch!=27)
       { if (*dos_filename<8)
             if (ch>47&&ch<58||ch>64&&ch<91||ch>96&&ch<123)
               { dos_filename[++*dos_filename]=(ch>96&&ch<123)?ch&(255-32):ch;
                 bar(397,45,462,55); // erase '--------' with white box
                 memcpy(info+10,dos_filename+1,8);
                 textxy(441,51,info,COUL_TEXT);
               }
          }

      if (ch!=27&&*dos_filename)
        { memcpy(dos_filename+*dos_filename+1,".SAV\0",5);
          bar(282,32,598,58); // efface zone texte
          if ((fp=fopen(dos_filename+1,"wb"))!=NULL)
            { fwrite(historique, sizeof(historique), 1, fp);
              textxy(440,39,"Game saved",COUL_TEXT);
              textxy(440,51,dos_filename+1,COUL_TEXT);
              fclose(fp);
            }
          else
            { textxy(440,39,"GAME NOT SAVED !",COUL_TEXT);
              textxy(440,51,"MAYBE INCORRECT FILENAME...",COUL_TEXT);
            }
          while(!kbhit()&&!mouse_bout()); while(mouse_bout());
        }
      mouse_curs(2);
      putimage(282, 32, bitmap, COPY_PUT);
      mouse_curs(1);
    }

  bouton(92,440,0,"LOAD");
  bouton(12,440,0,"SAVE");
  farfree(bitmap);
}

/*********************************** SOURIS *********************************/
// mouse events: int 0x33.

// The managemenf of those software interruptions is referenced in the book
// 'Bible du Pc'  (Micro Application)

void mouse_win(int minx, int maxx, int miny, int maxy)

{ Regs.x.ax = 0x07; // function 0x07
  Regs.x.cx = minx; // -> register cx
  Regs.x.dx = maxx; // -> register dx
  int86(0x33, &Regs, &Regs); // exec function 0x07, interrupt. 0x33

  Regs.x.ax = 0x08;
  Regs.x.cx = miny;
  Regs.x.dx = maxy;
  int86(0x33, &Regs, &Regs);
}

void mouse_curs(int show_hide_init)

{ Regs.x.ax = show_hide_init;
  int86(0x33, &Regs, &Regs);
}

int mouse_bout()

{ Regs.x.ax = 0x03;
  int86(0x33, &Regs, &Regs);
  return(Regs.x.bx);   // register bx ->
}

void mouse_pos(int *mouse_x, int *mouse_y)

{ Regs.x.ax = 0x03;
  int86(0x33, &Regs, &Regs);
  *mouse_x = Regs.x.cx;
  *mouse_y = Regs.x.dx;
}

void mouse_set(int x, int y)

{ Regs.x.ax = 0x04;
  Regs.x.cx = x;
  Regs.x.dx = y;
  int86(0x33, &Regs, &Regs);
}
